-- {{{ HEADER: File headers
module ToccCompile where

import AbsTocc
import PrintTocc
import ErrM
import List
import Maybe

import ToccProto
import ToccListVar
import ToccLocals
import ToccLocalsDecl
import ToccLib

-- }}}

-- {{{ DATA: CCode type declaration and functions

-- CCode data type consists of types (to go first in header file); headers (to
-- be placed after types in header files); and source, to be put in source file.
-- Example: To specify only source, use: defCCodecSrc = ".."}

data CCode = CCode  { cTypes, cSrc, cHeader :: String
                    } deriving (Eq,Ord,Show)

defCCode :: CCode
defCCode = CCode "" "" ""


-- concatenates CCode list, or errors in list if any.
catCCode :: [Err CCode] -> Err CCode
catCCode [] = Ok $ CCode "" "" ""
catCCode (c:cs) = case c of 
  Ok (CCode type1 src1 header1) -> case catCCode cs of
    Ok (CCode type2 src2 header2) -> Ok $ CCode (type1++type2) (src1++src2) (header1++header2)
    Bad s2 -> Bad s2
  Bad s1 -> case catCCode cs of 
    Ok _ -> Bad s1
    Bad s2 -> Bad (s1++s2)
 

-- }}}

-- {{{ DATA: ProcData type declaration and functions

-- ProcData contains information used by tplProc to construct c++ source
-- code from templates. To specify only some parameters, use: defProc{param1=..,
-- param2=..}

data ProcData = ProcData {      tdName                      :: String,
                                tdDescr                     :: String,
                                tdTypeName                  :: String,
                                tdUncertain                 :: Bool,
                                tdDiscovery                 :: Bool,
                                tdLocals                    :: Locals,
                                tdNewVariables              :: [Variable_spec],
                                tdExpandCode, tdRunCode     :: String,
                                tdCommChannel, tdCommVar    :: String,
                                tdExecTime                  :: Maybe Exp
                             } deriving (Eq,Ord,Show)


defProc :: ProcData
defProc = ProcData "" "" "" False False [] [] "" "" "" "" Nothing

-- }}}

-- {{{ FUNCTIONS: Helper functions 

assumeNoError :: Err a -> a
assumeNoError x = case x of
  Bad s -> error s
  Ok y -> y


intExp :: Integer -> Exp
intExp x = ExpConst $ ConstInt x 


-- Add underscored index to a string.
ixName :: String -> Int -> String
ixName name i = name ++ "_" ++ show i


-- Create a Local for a variable of C++ type "Proc" and name `name`.
localProc :: String -> Local
localProc name = Local QualReadWrite (typeCustom "Proc") ArrayNone IndexNone name ValueUnknown


-- Wraps the text into a fold, with foldmarks commented.
wrapInFold :: String -> String -> String
wrapInFold _ "" = ""
wrapInFold descr text = "// {{{ " ++ descr ++ "\n" ++ text ++ "// }}}\n"


-- Used as `intersperse', except that it does not add the separator between
-- empty elements.
sepNonEmpty :: String -> [String] -> String
sepNonEmpty _ [] = ""
sepNonEmpty sep (x:xs) = if null x || null rest
  then x ++ rest
  else x ++ sep ++ rest
  where rest = sepNonEmpty sep xs


-- }}}

-- {{{ FUNCTIONS: Functions to read parameters from AST elements

initVariable_spec :: Variable_spec -> Init_spec
initVariable_spec x = case x of 
  Var decl_spec init_spec  -> init_spec


ssblockProc :: Proc_spec -> SingleStBlock_spec
ssblockProc x = case x of
  ProcFor _ _ _ _ singlestblock_spec     -> singlestblock_spec
  ProcWhile _ singlestblock_spec   -> singlestblock_spec
  ProcIf _ singlestblock_spec      -> singlestblock_spec
  ProcElse singlestblock_spec    -> singlestblock_spec
  ProcElseIf _ singlestblock_spec  -> singlestblock_spec
  ProcTime _ singlestblock_spec    -> singlestblock_spec


expArg :: Arg_spec -> Exp
expArg x = case x of
  Arg exp -> exp


-- }}}

-- {{{ FUNCTIONS: Compiling types

isChannel :: Qualifier_spec -> Bool
isChannel x = case x of
  QualChannelOut -> True
  QualChannelIn -> True
  QualChannel -> True
  _ -> False


compileType_spec :: Type_spec -> String
compileType_spec x = case x of
  TypeString  -> "std::string"
  TypeBool  -> "bool"
  TypeInt  -> "int"
  TypeFloat  -> "double"
  TypeCustom upperident  -> (printTree upperident) 


compileArray_spec :: Qualifier_spec -> Array_spec -> String -> String
compileArray_spec qualifier_spec array_spec typeName = case array_spec of
  ArraySized _  -> "std::vector<" ++ typeName ++ (if isChannel qualifier_spec then "*" else "") ++ ">"
  ArrayUnknown ->  "std::vector<" ++ typeName ++ (if isChannel qualifier_spec then "*" else "") ++ ">"
  ArrayNone  -> typeName


compileQualifier_spec :: Qualifier_spec -> String -> String
compileQualifier_spec x typeName = case x of
  QualChannel -> "Channel<" ++ typeName ++ "> "
  QualChannelIn -> "ReceiveChannel<" ++ typeName ++ "> "
  QualChannelOut -> "SendChannel<" ++ typeName ++ "> "
  QualByVal -> typeName
  QualConst -> "const " ++ typeName
  QualReadWrite -> typeName


compileType :: Local -> String
compileType x = case x of
  Local QualChannel type_spec (ArraySized _) _ name _ -> "ChannelVector<" ++ compileType_spec (locType x) ++ ">"
  Local qualifier_spec type_spec array_spec indexing name _ -> 
    compileArray_spec qualifier_spec array_spec $
    compileQualifier_spec qualifier_spec $
    compileType_spec type_spec


-- }}}

-- {{{ FUNCTIONS: Compile argument, member and parameter lists

-- New class argument list, from existing scope locals. This converts between
-- ways to pass arguments as different types of parameters, adding
-- dereference operators, and makeSend/makeReceive.
arglLocal :: (Local, Local, String) -> Err String
arglLocal (param, arg, name) = case (locQualifier param, locQualifier arg) of
  (QualChannel, QualChannel) -> Ok $ name
  (QualChannelIn, QualChannel) -> Ok $ "makeReceive(" ++ name ++ ")"
  (QualChannelIn, QualChannelIn) -> Ok $ name
  (QualChannelOut, QualChannel) -> Ok $ "makeSend(" ++ name ++ ")"
  (QualChannelOut, QualChannelOut) -> Ok $ name
  (QualByVal, QualByVal) -> Ok $ name
  (QualConst, QualConst) -> Ok $ name
  (QualConst, QualReadWrite) -> Ok $ name
  (QualReadWrite, QualReadWrite) -> Ok $ name
  (qp, qa) -> Bad $ "Argument qualifier `" ++ printTree qa ++ "' does not match parameter qualifier `" ++ printTree qp ++ "'"
  -- _ -> Bad $ "\nParam: " ++ show param ++ "\nArg: " ++ show arg ++ "\nName: " ++ show name


-- Vectors of channels are passed as entire vectors, but the Locals list
-- may include individual elements. Duplicate vector elements are nubbed away.
arglLocals :: String -> Locals -> Locals -> String
arglLocals pop [] args = sepNonEmpty ", " $ pop : (map locName args)
arglLocals pop params args = 
  let (ps, as) = unzip $ nubBy (\x y -> fst x == fst y) (zip params args) in
    sepNonEmpty ", " (pop : (map (assumeNoError.arglLocal) $ zip3 ps as (map locName as)))


-- Class constructor initializer list, from existing scope locals.  Parameters
-- taken by value are converted to pointers, others are passed as they are.
initlLocal :: Local -> String
initlLocal x = if (isChannel.locQualifier) x && locArray x == ArrayNone
    then locName x ++ "(take(P_" ++ locName x ++ "))"
    else locName x ++ "(P_" ++ locName x ++ ")"
  

initlLocals :: Locals -> String
initlLocals xs = sepNonEmpty ",\n" $ map initlLocal xs


-- Class constructor initializer list, from new scope variables
initlNewVariable :: Locals -> Variable_spec -> String
initlNewVariable locals x = locName l ++ "(new " ++ compileType l ++ "(" ++ init ++ "))"
  where
    l = localVariable_spec [] x
    init = case (locArray l, initVariable_spec x) of
      (ArraySized expSize, InitExp expInit) -> 
        compileExp locals expSize ++ ", " ++ compileExp locals expInit
      (ArraySized expSize, InitNone) -> compileExp locals expSize
      (ArrayNone, InitExp expInit) -> compileExp locals expInit
      _ -> ""
      

initlNewVariables :: Locals -> [Variable_spec] -> String
initlNewVariables locals xs = sepNonEmpty ",\n" $ map (initlNewVariable locals) xs


-- Class parameter list from existing locals
paramlLocal :: Local -> String
paramlLocal x = case locQualifier x of
  QualByVal -> compileType x ++ " P_" ++ locName x
  _ -> compileType x ++ "* P_" ++ locName x


paramlLocals :: Locals -> String
paramlLocals xs = sepNonEmpty ", " $ "Proc* pop":map paramlLocal xs


-- Class member list from existing locals
memberlLocal :: Local -> String
memberlLocal x = if locQualifier x == QualByVal
  then compileType x ++ " " ++ locName x ++ ";\n"
  else compileType x ++ "* " ++ locName x ++ ";\n"


memberlLocals :: Locals -> String
memberlLocals xs = concat $ map memberlLocal xs


-- Class member list from new scope variables
memberlVariable_specs :: [Variable_spec] -> String
memberlVariable_specs xs = memberlLocals $ localsVariable_specs [] xs


-- }}}

-- {{{ FUNCTIONS: Compile argument lists for calling named processes

callarglByRef :: Locals -> Local -> String -> String -> Err String
callarglByRef locals param name text = case find ((==name).locName) locals of
  Nothing -> Bad $ "Variable `" ++ name ++ "' not found"
  Just l -> arglLocal (param, l, text)


callarglExp :: Locals -> Local -> Exp -> Err String
callarglExp locals param x = case x of
  ExpIndex lowerident exp -> let id = printTree lowerident in
    if isChannel (locQualifier param)
      then callarglByRef locals param id ("(*" ++ id ++ ")[" ++ compileExp locals exp ++ "]")
      else callarglByRef locals param id ("&(*" ++ id ++ ")[" ++ compileExp locals exp ++ "]")
  ExpVar lowerident -> 
    callarglByRef locals param (printTree lowerident) (printTree lowerident)
  _ -> Bad $ "Cannot pass expression by reference"


callarglArg_spec :: Program_spec -> Locals -> String -> (Int, Arg_spec) -> Err String
callarglArg_spec ps locals proc (i,x) = case lookupProto ps proc i of
  Bad s -> Bad s
  Ok p -> case locQualifier p of 
    QualByVal -> Ok $ compileExp locals (expArg x)
    _ -> case callarglExp locals p (expArg x) of 
      Bad s -> Bad $ s ++ ", paramter #" ++ show i ++ " when calling process `" ++ proc ++ "'\n"
      Ok a -> Ok a


callarglArg_specs :: Program_spec -> String -> Locals -> [Arg_spec] -> Err String
callarglArg_specs ps proc locals args = do
  a <- mapM (callarglArg_spec ps locals proc) (zip [1..] args)
  return (concat $ intersperse ", " $ "pop":a)


-- }}}

-- {{{ FUNCTIONS: Transform expressions (add '*' to pointer variables)

expVar :: Locals -> LowerIdent -> String
expVar locals lowerident = let name = printTree lowerident in 
  case find ((==name).locName) locals of
    Just (Local {locQualifier=QualByVal}) -> name
    Just _ -> "(*" ++ name ++ ")"
    Nothing -> if "P_" `isPrefixOf` name
      then name
      else "(*" ++ name ++ ")"


compileExp :: Locals -> Exp -> String
compileExp locals x = case x of
  ExpLor exp0 exp      -> "(" ++ (compileExp locals exp0) ++ "||" ++ (compileExp locals exp) ++ ")"
  ExpLand exp0 exp     -> "(" ++ (compileExp locals exp0) ++ "&&" ++ (compileExp locals exp) ++ ")"
  ExpBitor exp0 exp    -> "(" ++ (compileExp locals exp0) ++ "|" ++ (compileExp locals exp) ++ ")"
  ExpBitexor exp0 exp  -> "(" ++ (compileExp locals exp0) ++ "^" ++ (compileExp locals exp) ++ ")"
  ExpBitand exp0 exp   -> "(" ++ (compileExp locals exp0) ++ "&" ++ (compileExp locals exp) ++ ")"
  Expeq exp0 exp       -> "(" ++ (compileExp locals exp0) ++ "==" ++ (compileExp locals exp) ++ ")"
  Expneq exp0 exp      -> "(" ++ (compileExp locals exp0) ++ "!=" ++ (compileExp locals exp) ++ ")"
  Explthen exp0 exp    -> "(" ++ (compileExp locals exp0) ++ "<" ++ (compileExp locals exp) ++ ")"
  Expgrthen exp0 exp   -> "(" ++ (compileExp locals exp0) ++ ">" ++ (compileExp locals exp) ++ ")"
  Exple exp0 exp       -> "(" ++ (compileExp locals exp0) ++ "<=" ++ (compileExp locals exp) ++ ")"
  Expge exp0 exp       -> "(" ++ (compileExp locals exp0) ++ ">=" ++ (compileExp locals exp) ++ ")"
  ExpLeft exp0 exp     -> "(" ++ (compileExp locals exp0) ++ "<<" ++ (compileExp locals exp) ++ ")"
  ExpRight exp0 exp    -> "(" ++ (compileExp locals exp0) ++ ">>" ++ (compileExp locals exp) ++ ")"
  ExpPlus exp0 exp     -> "(" ++ (compileExp locals exp0) ++ "+" ++ (compileExp locals exp) ++ ")"
  ExpMinus exp0 exp    -> "(" ++ (compileExp locals exp0) ++ "-" ++ (compileExp locals exp) ++ ")"
  ExpTimes exp0 exp    -> "(" ++ (compileExp locals exp0) ++ "*" ++ (compileExp locals exp) ++ ")"
  ExpDiv exp0 exp      -> "(" ++ (compileExp locals exp0) ++ "/" ++ (compileExp locals exp) ++ ")"
  ExpMod exp0 exp      -> "(" ++ (compileExp locals exp0) ++ "%" ++ (compileExp locals exp) ++ ")"
  ExpNegative exp      -> "(" ++ " -" ++ (compileExp locals exp) ++ ")"
  ExpPositive exp      -> "(" ++ " +" ++ (compileExp locals exp) ++ ")"
  ExpIndex lowerident exp     -> "(*" ++ (printTree lowerident) ++ ")[" ++ (compileExp locals exp) ++ "]"
  ExpFunc lowerident exps     -> 
    "(" ++ (printTree lowerident) ++ "(" ++ (concat $ intersperse ", " $ map (compileExp locals) exps) ++ "))"
  c@(ExpConst constval_spec)  -> printTree c
  ExpVar lowerident           -> expVar locals lowerident

-- }}}

-- {{{ FUNCTIONS: Conditional statement (IF/ELSE/ELSEIF)

-- Convert from ProcIf/ProcElseIf/ProcElse to C-Code conditionals.
compileConditionals :: Locals -> String -> Proc_spec -> String
compileConditionals locals name x = case x of
  ProcIf exp singlestblock_spec -> "\tif (" ++ (compileExp locals exp) ++ ") {\n\t" ++ connect ++ "}\n"
  ProcElseIf exp singlestblock_spec -> "\telse if (" ++ (compileExp locals exp) ++ ") {\n\t" ++ connect ++ "}\n"
  ProcElse singlestblock_spec -> "\telse {\n\t" ++ connect ++ "}\n"
  where
    connect = tplConnect name "pop" [] locals


-- Implementing an expand() function disables the superclass default expand(),
-- and if there is no explicit "else", this default expand must be added.
compileDefaultElse :: [Proc_spec] -> String
compileDefaultElse xs = case last xs of
  ProcElse _ -> ""
  _ ->  "\telse\n\t\tconnect(pop);\n"
  

-- Convert list of if/elseif*/else? to C-code.
compileIf :: Program_spec -> String -> Locals -> [Proc_spec] -> Err CCode
compileIf ps name locals procs = catCCode [ifCode, thenCode]
  where
    ifStmt (i,p) = compileConditionals locals (ixName name i) p 
    thenStmt (i,p) = compileSingleStBlock_spec ps locals (ixName name i) (ssblockProc p) 
    expandCode = (concatMap ifStmt $ zip [1..] procs) ++ compileDefaultElse procs 
    ifCode = tplProc defProc {tdName=name, tdTypeName="ControlProc", tdDescr="If", tdLocals=locals, 
      tdExpandCode=expandCode, tdUncertain=True, tdDiscovery=False}
    thenCode = catCCode $ map thenStmt (zip [1..] procs)


-- }}}

-- {{{ FUNCTIONS: Compiling new scope

-- Adds a scope layer class, adding local variables and removing conflicting
-- ones.
compileScope :: Program_spec -> String -> Locals -> [Variable_spec] -> (Err CCode, Locals)
compileScope ps name locals vars = (catCCode [codeNewScope, codeDelScope], newLocals)
  where
    childName = ixName name 1
    delName = ixName name 2
    addLocals = localsVariable_specs locals vars
    newLocals = nestScope locals addLocals
    expandCode = tplTmpProc delName "pop" newLocals ++ tplConnect childName "pop" [] newLocals 
    runCode = concat $ map (\l -> "delete " ++ locName l ++ ";\n") addLocals
    codeNewScope = tplProc defProc {tdName=name, tdTypeName="ControlProc",
      tdDescr="New Scope", tdLocals=locals, tdNewVariables=vars,
      tdExpandCode=expandCode, tdDiscovery=True}
    codeDelScope = tplProc defProc {tdName=delName, tdTypeName="CodeProc", 
      tdDescr="Delete Scope", tdLocals=newLocals, tdRunCode=runCode, tdExecTime=Just $ intExp 1}


-- }}}

-- {{{ FUNCTIONS: Compile code proc

compileCode :: Program_spec -> Locals -> String -> String -> String -> Exp -> String -> Err CCode
compileCode ps locals descr name next time code = tplProc defProc {tdName=name,
  tdTypeName="CodeProc", tdDescr=descr, tdLocals=locals,
  tdExpandCode=expandCode, tdRunCode=code, tdExecTime=Just time}
  where
    expandCode = tplConnect next "pop" [] locals 


-- }}}

-- {{{ FUNCTIONS: WHILE

compileWhile :: Program_spec -> Locals -> String -> String -> Exp -> SingleStBlock_spec -> Err CCode
compileWhile ps locals name next exp ssblock = catCCode [whileCode, blockCode]
  where
    expandCode = "if (" ++ compileExp locals exp ++ ") {\n"
      ++ indent "\t" (
         tplTmpProc name "loop" locals 
      ++ tplConnect (ixName name 1) "loop" [] locals
      )
      ++ "}\n"
      ++ "else {\n"
      ++ indent "\t" (
         tplTmpProc next "pop" locals
      ++ "connect(pop);\n"
      )
      ++ "}\n"
    whileCode = tplProc defProc {tdName=name, tdTypeName="ControlProc", tdDescr="While",
      tdLocals=locals, tdExpandCode=expandCode, tdDiscovery=True}
    blockCode = compileSingleStBlock_spec ps locals (ixName name 1) ssblock

-- }}}

-- {{{ FUNCTIONS: SEQ, PAR, SingleStBlock_spec

compileSingleProcs :: Program_spec -> Locals -> String -> [Proc_spec] -> Err CCode
compileSingleProcs ps locals name proc_specs = case partitionProcs proc_specs of
  [conditional@((ProcIf _ _):_)] -> compileIf ps name locals conditional
  [[proc]] -> compileProc_spec ps locals name "" proc
  _ -> Bad "Single statement, or ordered multiple statements expected (SEQ/PAR).\n"


compileSingleStBlock_spec :: Program_spec -> Locals -> String -> SingleStBlock_spec -> Err CCode
compileSingleStBlock_spec ps locals name x = case x of
  SingleStBlock variable_specs proc_specs  -> if null variable_specs
    then compileSingleProcs ps locals name proc_specs
    else catCCode [scopeCode, blockCode] 
    where
      (scopeCode, newLocals) = compileScope ps name locals variable_specs
      blockCode = compileSingleProcs ps newLocals (ixName name 1) proc_specs


compileMultiProcs :: Program_spec -> Locals -> String -> String -> Order_spec -> [Proc_spec] -> Err CCode
compileMultiProcs ps locals name next order proc_specs = case order of 
  OrderSeq -> compileSeq ps locals name next 1 proc_specs
  OrderPar -> compilePar ps locals name next proc_specs


compileOrder :: Program_spec -> Locals -> String -> String -> Order_spec -> Block_spec -> Err CCode
compileOrder ps locals name next order_spec block_spec = case block_spec of
  Block variable_specs proc_specs  -> if null variable_specs
    then compileMultiProcs ps locals name next order_spec proc_specs
    else catCCode [scopeCode, blockCode] 
    where
      (scopeCode, newLocals) = compileScope ps name locals variable_specs
      blockCode = compileMultiProcs ps locals (ixName name 1) next order_spec proc_specs


compileSeq :: Program_spec -> Locals -> String -> String -> Int -> [Proc_spec] -> Err CCode
compileSeq _ _ _ _ _ [] = Ok defCCode
compileSeq ps locals name next ix (x:xs) = let 
  thisName = if ix == 1 then name else ixName name ix
  nextName = if null xs then next else (ixName name (ix+1)) 
  in case x of
    ProcIf _ _ -> catCCode [ifCode, restCode] where
      ifCode = compileIf ps thisName locals (x:(fst $ partitionElseIfs xs))
      restCode = compileSeq ps locals name next (ix+1) (snd $ partitionElseIfs xs)
    _ -> catCCode [thisCode, restCode] where
      thisCode = compileProc_spec ps locals thisName nextName x
      restCode = compileSeq ps locals name next (ix+1) xs


compileParExpand :: String -> (Int, (Locals, Locals)) -> String
compileParExpand basename (index, (params, args)) = tplConnect (ixName basename index) "join" params args 


compilePar :: Program_spec -> Locals -> String -> String -> [Proc_spec] -> Err CCode
compilePar ps locals name next procs = let 
  ess = mapM (needLocals ps locals) procs
  in case ess of
    Bad s -> Bad s
    Ok es -> catCCode [parCode, restCode]
      where
        expandCode = tplTmpProc next "pop" locals
          ++ "ParEndProc* join = add(new ParEndProc(pop, this));\n"
          ++ concatMap (compileParExpand name) (zip [1..] es)
        parCode = tplProc defProc {tdName=name, tdTypeName="ParProc", tdDescr="PAR", tdLocals=locals, 
          tdExpandCode=expandCode}
        restCode = catCCode $ map (\(i,x,(p,_)) -> compileProc_spec ps p (ixName name i) "" x) (zip3 [1..] procs es)


-- }}}

-- {{{ FUNCTIONS: Replication: SEQ FOR, PAR FOR

compileSeqFor :: Program_spec -> Locals -> String -> String -> LowerIdent -> Exp -> Exp -> SingleStBlock_spec -> Err CCode
compileSeqFor ps locals name next lowerident exp0 exp ssblock = catCCode [initCode, forCode, blockCode] where
  ctr = printTree lowerident
  newLocal = localsIterVar lowerident ValueIterator
  forLocals = nestScope locals [newLocal]
  initExpandCode = "connect(add(new " ++ (ixName name 1) ++ "(" ++ arglLocals "pop" [] locals ++ ", " 
    ++ compileExp locals exp0 ++ ")));\n"
  forExpandCode = "if (" ++ ctr ++ " <= " ++ compileExp locals exp ++ ") {\n"
    ++ "\tProc* pop = add(new " ++ (ixName name 1) ++ "(" ++ arglLocals "this->pop" [] locals ++ ", "
    ++ ctr ++ "+1));\n"
    ++ "\t" ++ tplConnect (ixName name 2) "pop" [] forLocals
    ++ "\t" ++ ctr ++ "++;\n"
    ++ "}\n"
    ++ "else\n"
    ++ "\tconnect(pop);\n"
  initCode = tplProc defProc {tdName=name, tdTypeName="ControlProc", tdDescr="SEQ FOR INIT",
    tdLocals=locals, tdExpandCode=initExpandCode, tdDiscovery=True}
  forCode = tplProc defProc {tdName=ixName name 1, tdTypeName="ControlProc", tdDescr="SEQ FOR",
    tdLocals=forLocals, tdExpandCode=forExpandCode, tdDiscovery=True}
  blockCode = compileSingleStBlock_spec ps forLocals (ixName name 2) ssblock


compileParFor :: Program_spec -> Locals -> String -> String -> LowerIdent -> Exp -> Exp -> SingleStBlock_spec -> Err CCode
compileParFor ps locals name next lowerident exp0 exp ssblock = catCCode [forCode, blockCode] where
  ctr = printTree lowerident
  newVar = (localsIterVar lowerident ValueIterator){locQualifier=QualByVal}
  newLocals = nestScope locals [newVar]
  expandCode = tplTmpProc next "pop" locals
    ++ "ParEndProc* join = add(new ParEndProc(pop, this));\n"
    ++ "int " ++ ctr ++ ";\n"
    ++ "for (" ++ ctr ++ " = " ++ compileExp locals exp0 ++ "; " 
    ++ ctr ++ " <= " ++ compileExp locals exp ++ "; (" ++ ctr ++ ")++) {\n"
    ++ "\t" ++ tplConnect (ixName name 1) "join" [] newLocals
    ++ "}\n"
  forCode = tplProc defProc {tdName=name, tdTypeName="ParProc", tdDescr="PAR FOR", 
    tdLocals=locals, tdExpandCode=expandCode}
  blockCode = compileSingleStBlock_spec ps newLocals (ixName name 1) ssblock


compileFor :: Program_spec -> Locals -> String -> String -> Order_spec -> LowerIdent -> Exp -> Exp -> SingleStBlock_spec -> Err CCode
compileFor ps locals name next order_spec lowerident exp0 exp = case order_spec of
  OrderSeq -> compileSeqFor ps locals name next lowerident exp0 exp
  OrderPar -> compileParFor ps locals name next lowerident exp0 exp


-- }}} 

-- {{{ FUNCTIONS: ALT functions
  
-- _phase is the iterator in a for-loop from 0 to 2. 0 means that only ready
-- channel alternatives will be accepted. 1 means that skips will be accepted.
-- 2 means that anything goes.

guardCodeAltAlt_specs :: Locals -> Maybe Exp -> AltComm_spec -> String
guardCodeAltAlt_specs locals mexp x = case x of
  AltCommRecv lowerident0 lowerident -> let ch = printTree lowerident0 in case mexp of 
    Nothing -> "if (_phase == 2 || " ++ ch ++ "->getChannelBase()->isSenderReady()) "
    Just exp -> "if (" ++ compileExp locals exp ++ " && (_phase == 2 || " ++ ch ++ "->getChannelBase()->isSenderReady()))"


-- expandCode is the code that selects alternative.
expandCodeAltAlt_specs :: Program_spec -> Locals -> String -> Int -> [AltAlt_spec] -> String
expandCodeAltAlt_specs _ _ _ _ [] = ""
expandCodeAltAlt_specs ps locals name ix (x:xs) = concat [this, rest]
  where
    this = case x of
      AltAltNested alt_spec  -> expandCodeAlt_spec ps locals (ixName name ix) alt_spec
      AltAltGuard exp altcomm_spec _  -> guardCodeAltAlt_specs locals (Just exp) altcomm_spec ++ "{\n\t"
        ++ tplConnect (ixName name ix) "pop" [] locals
        ++ "\treturn;\n"
        ++ "}\n"
      AltAltSkip exp  -> "if (_phase == 1 && " ++ compileExp locals exp ++ ") {\n"
        ++ "\tconnect(pop);\n"
        ++ "\treturn;\n"
        ++ "}\n"
      AltAlt altcomm_spec _  -> guardCodeAltAlt_specs locals Nothing altcomm_spec ++ "{\n\t"
        ++ tplConnect (ixName name ix) "pop" [] locals
        ++ "\treturn;\n"
        ++ "}\n"
    rest = expandCodeAltAlt_specs ps locals name (ix+1) xs


expandCodeAlt_spec :: Program_spec -> Locals -> String -> Alt_spec -> String
expandCodeAlt_spec ps locals name x = case x of
  Alt altalt_specs  -> expandCodeAltAlt_specs ps locals name 1 altalt_specs
  AltRepl lowerident exp0 exp altalt_specs  -> let 
    ctr = printTree lowerident 
    newLocals = nestScope locals [localsIterVar lowerident ValueKnownExpandTime]
    in
      "for (int" ++ ctr ++ " = " ++ compileExp locals exp0 ++ "; "
      ++ ctr ++ " <= " ++ compileExp locals exp ++ "; ++" ++ ctr ++ ") {\n"
      ++ (indent "\t" $ expandCodeAltAlt_specs ps newLocals name 1 altalt_specs)


-- blockCode is the action to be done for the alternative.
blockCodeAltComm_spec :: Program_spec -> Locals -> String -> String -> AltComm_spec -> Err CCode
blockCodeAltComm_spec ps locals name next x = case x of
  AltCommRecv lowerident0 lowerident  -> compileComm_spec ps locals name next $ CommRecv lowerident0 lowerident


blockCodeAltAlt_specs :: Program_spec -> Locals -> String -> Int -> [AltAlt_spec] -> Err CCode
blockCodeAltAlt_specs _ _ _ _ [] = Ok defCCode
blockCodeAltAlt_specs ps locals name ix (x:xs) = catCCode [this, rest] where
  thisName = ixName name ix
  childName = ixName thisName 1
  this = case x of
    AltAltNested alt_spec  -> blockCodeAlt_spec ps locals thisName alt_spec
    AltAltGuard _ altcomm_spec singlestblock_spec  -> 
      catCCode [blockCodeAltComm_spec ps locals thisName childName altcomm_spec, 
      compileSingleStBlock_spec ps locals childName singlestblock_spec]
    AltAltSkip _  -> Ok defCCode
    AltAlt altcomm_spec singlestblock_spec  -> 
      catCCode [blockCodeAltComm_spec ps locals thisName childName altcomm_spec, 
      compileSingleStBlock_spec ps locals childName singlestblock_spec]
  rest = blockCodeAltAlt_specs ps locals name (ix+1) xs


blockCodeAlt_spec :: Program_spec -> Locals -> String -> Alt_spec -> Err CCode
blockCodeAlt_spec ps locals name x = case x of
  Alt altalt_specs  -> blockCodeAltAlt_specs ps locals name 1 altalt_specs
  AltRepl lowerident exp0 exp altalt_specs  -> blockCodeAltAlt_specs ps locals name 1 altalt_specs


compileAlt_spec :: Program_spec -> Locals -> String -> String -> Alt_spec -> Err CCode
compileAlt_spec ps locals name next x = catCCode [altCode, blockCode] where
  expandCode = tplTmpProc next "pop" locals 
    ++ "for (int _phase = 0; _phase <= 2; _phase++) {\n"
    ++ (indent "\t" $ expandCodeAlt_spec ps locals name x)
    ++ "}\n"
  altCode = tplProc defProc {tdName=name, tdTypeName="ControlProc", tdDescr="ALT", 
    tdLocals=locals, tdExpandCode=expandCode}
  blockCode = blockCodeAlt_spec ps locals name x


-- }}}

-- {{{ FUNCTIONS: Compile top levels: Process, Include, Typedef, Global_specs

wrapProcess :: CapIdent -> Err CCode -> Err CCode
wrapProcess capident x = case x of
  Bad s -> Bad $ "In function `" ++ name ++ "':\n" ++ indent " " s
  Ok (CCode {cSrc=src, cHeader=header}) -> Ok defCCode {
    cSrc = wrapInFold ("Source for process `" ++ name ++ "'") src ++ "\n",
    cHeader = wrapInFold ("Headers for process `" ++ name ++ "'") header ++ "\n"}
  where
    name = printTree capident


compileProcess :: Program_spec -> CapIdent -> [Param_spec] -> SingleStBlock_spec -> Err CCode
compileProcess ps capident param_specs singlestblock_spec =  wrapProcess capident $ catCCode [mainCode, stmCode] 
  where
    name = printTree capident
    childName = ixName name 1
    locals = localsParam_specs param_specs
    stmCode = compileSingleStBlock_spec ps locals childName singlestblock_spec
    mainCode = tplProc defProc {tdName=name, tdTypeName="ControlProc",
      tdDescr="Process", tdLocals=locals, tdExpandCode=tplConnect (ixName name 1) "pop" [] locals, 
      tdDiscovery=True}


compileSInclude :: String -> Err CCode
compileSInclude s = Ok defCCode {cTypes="\n#include <" ++ s ++ ">\n"}


compileInclude :: String -> Err CCode
compileInclude s = Ok defCCode {cTypes="\n#include \"" ++ s ++ "\"\n"}


compileTypedef :: UpperIdent -> Type_spec -> Array_spec -> Err CCode
compileTypedef upperident type_spec array_spec = Ok defCCode {cTypes=code}
  where
    code = "typedef " ++ printTree type_spec ++ " " ++ printTree upperident 
      ++ " " ++ printTree array_spec ++ ";\n" 
  

compileGlobal_specs :: Program_spec -> [Global_spec] -> Err CCode
compileGlobal_specs ps xs = catCCode $ map (compileGlobal_spec ps) xs


compileGlobal_spec :: Program_spec -> Global_spec -> Err CCode
compileGlobal_spec ps x = case x of
  GlobalProcess capident param_specs singlestblock_spec -> compileProcess ps capident param_specs singlestblock_spec
  GlobalTypedef upperident type_spec array_spec -> compileTypedef upperident type_spec array_spec
  GlobalInclude str  -> compileInclude str
  GlobalSInclude str  -> compileInclude str
      


-- }}}

-- {{{ FUNCTIONS: Other parts of syntax tree: Comm/Delay/Exp/Print/Stop/Skip/Time/Assign/Call
 
compileComm_spec :: Program_spec -> Locals -> String -> String -> Comm_spec -> Err CCode
compileComm_spec ps locals name next x = case x of
  CommSend lowerident0 lowerident  -> commProc (printTree lowerident0) (printTree lowerident)
  CommRecv lowerident0 lowerident  -> commProc (printTree lowerident0) (printTree lowerident)
  where
    expandCode = tplConnect next "pop" [] locals
    commProc ch var = tplProc defProc {tdName=name, tdTypeName="CommProc", tdDescr="Comm", 
      tdLocals=locals, tdExpandCode=expandCode, tdCommChannel=ch, tdCommVar=var}


compileDelay :: Program_spec -> Locals -> String -> String -> Exp -> Err CCode
compileDelay ps locals name next exp = compileCode ps locals "DELAY" name next exp $
  "usleep(" ++ compileExp locals exp ++ ");\n"
 

-- TODO: Exectime at 1 is completely arbitrary.
compileExpProc :: Program_spec -> Locals -> String -> String -> Exp -> Err CCode
compileExpProc ps locals name next exp = compileCode ps locals "(expr)" name next (intExp 1) $
  compileExp locals exp ++ ";\n"
 

compilePrint :: Program_spec -> Locals -> String -> String -> [Exp] -> Err CCode
compilePrint ps locals name next exps = compileCode ps locals "PRINT" name next (intExp 1) $
  "std::cout << " ++ concat (intersperse " << " (map (compileExp locals) exps)) ++ " << std::endl;\n"
 

compileStop :: Program_spec -> Locals -> String -> String -> Err CCode
compileStop ps locals name next = tplProc defProc {tdName=name,
  tdTypeName="ControlProc", tdDescr="STOP", tdLocals=locals,
  tdExpandCode="\n", tdDiscovery=False}
 

compileSkip :: Program_spec -> Locals -> String -> String -> Err CCode
compileSkip ps locals name next = tplProc defProc {tdName=name,
  tdTypeName="ControlProc", tdDescr="SKIP", tdLocals=locals,
  tdExpandCode=tplConnect next "pop" [] locals, tdDiscovery=False}


compileTime :: Program_spec -> Locals -> String -> String -> Exp -> SingleStBlock_spec -> Err CCode
compileTime ps locals name next exp ssblock = catCCode [timerCode, blockCode]
  where
    expandCode = tplTmpProc next "pop" locals
      ++ "TimerEndProc* pop = add(new TimerEndProc(this->pop));\n"
      ++ tplTmpProc (ixName name 1) "next" locals
      ++ "connect(add(new TimerProc(next, pop, " ++ (compileExp locals exp) ++ ")));\n"
    timerCode = tplProc defProc {tdName=name, tdTypeName="ControlProc", tdDescr="Timer", 
      tdLocals=locals, tdExpandCode=expandCode, tdDiscovery=True}
    blockCode = compileSingleStBlock_spec ps locals (ixName name 1) ssblock


compileAssign :: Program_spec -> Locals -> String -> String -> LowerIdent -> Assignment_op -> Exp -> Err CCode
compileAssign ps locals name next lowerident assignment_op exp = 
  compileCode ps locals "Assign" name next (intExp 1) 
  (rhs ++ " " ++ printTree assignment_op ++ " " ++ compileExp locals exp ++ ";\n")
  where rhs = "(*" ++ printTree lowerident ++ ")"
 

compileAssignArr :: Program_spec -> Locals -> String -> String -> LowerIdent -> Exp -> Assignment_op -> Exp -> Err CCode
compileAssignArr ps locals name next lowerident exp0 assignment_op exp = 
  compileCode ps locals "Assign" name next (intExp 1) 
  (printTree lowerident ++ "[" ++ compileExp locals exp0 ++ "] " ++ 
   printTree assignment_op ++ " " ++ compileExp locals exp ++ ";\n")


compileCall :: Program_spec -> Locals -> String -> String -> CapIdent -> [Arg_spec] -> Err CCode
compileCall ps locals name next capident arg_specs = 
  case callarglArg_specs ps (printTree capident) locals arg_specs of
    Bad s -> Bad s
    Ok argl -> tplProc defProc {tdName=name, tdTypeName="ControlProc", 
      tdDescr="Call", tdLocals=locals, tdExpandCode=expandCode, tdDiscovery=True}
      where
        expandCode = tplTmpProc next "pop" locals
          ++ "connect(add(new " ++ (printTree capident) ++ "(" ++ argl ++ ")));\n"


-- }}}

-- {{{ FUNCTIONS: compileProc_spec, distributing function.

compileProc_spec :: Program_spec -> Locals -> String -> String -> Proc_spec -> Err CCode
compileProc_spec ps locals name next x = case x of
  ProcDelay exp  -> compileDelay ps locals name next exp
  ProcExp exp  -> compileExpProc ps locals name next exp
  ProcFor order_spec lowerident exp0 exp singlestblock_spec  -> 
    compileFor ps locals name next order_spec lowerident exp0 exp singlestblock_spec
  ProcWhile exp singlestblock_spec  -> compileWhile ps locals name next exp singlestblock_spec
  ProcAlt alt_spec  -> compileAlt_spec ps locals name next alt_spec
  ProcElse singlestblock_spec  -> Bad "Stray ELSE in program\n"
  ProcElseIf exp singlestblock_spec  -> Bad "Stray ELSE IF in program\n"
  ProcPrint exps  -> compilePrint ps locals name next exps
  ProcSkip  -> compileSkip ps locals name next
  ProcStop  -> compileStop ps locals name next
  ProcOrder order block_spec  -> compileOrder ps locals name next order block_spec
  ProcTime exp singlestblock_spec  -> compileTime ps locals name next exp singlestblock_spec
  ProcAssign lowerident assignment_op exp  -> 
    compileAssign ps locals name next lowerident assignment_op exp
  ProcAssignArr lowerident exp0 assignment_op exp  -> 
    compileAssignArr ps locals name next lowerident exp0 assignment_op exp
  ProcComm comm_spec  -> compileComm_spec ps locals name next comm_spec
  ProcCall capident arg_specs  -> compileCall ps locals name next capident arg_specs


-- }}}

-- {{{ FUNCTIONS: Compiler entry functions

compileProgram_spec :: Err Program_spec -> Err CCode
compileProgram_spec (Bad s) = Bad s
compileProgram_spec (Ok x) = case x of
  Program global_specs -> compileGlobal_specs x global_specs


-- }}}

-- {{{ TEMPLATES: C++ class and function skeletons

tplProc :: ProcData -> Err CCode 
tplProc td = case td of
  ProcData name descr typeName uncertain discovery locals newVariables expandCode runCode commChannel commVar execTime -> 
    Ok $ defCCode {cSrc=src,cHeader=header}
    where
      commSuperClassParams = if null commChannel then "" else "P_" ++ commChannel ++ ", P_" ++ commVar
      superClassParams = ["pop", commSuperClassParams]
      ilSuperClass = typeName ++ "(" ++ (sepNonEmpty ", " superClassParams) ++ ")"
      ilLocals = initlLocals (filterScope locals $ localsVariable_specs [] newVariables)
      ilNewVariables = initlNewVariables locals newVariables
      initList = sepNonEmpty ",\n" [ilSuperClass, ilLocals, ilNewVariables]

      header = wrapInFold ("Class " ++ name ++ " : " ++ typeName ++ " (" ++ descr ++ ")") $
        "class " ++ name ++ " : public " ++ typeName ++ " {\n" 
        ++ indent "\t" (
           "private:\n" 
        ++ indent "\t" (
           memberlLocals (filterScope locals $ localsVariable_specs [] newVariables) 
        ++ memberlVariable_specs newVariables
        ) 
        ++ "public:\n"
        ++ indent "\t" (
           (if execTime /= Nothing then "virtual int getExecTime() const {return " 
            ++ (compileExp locals $ fromJust execTime) ++ "; }\n" else "")
        ++ (if not (null commChannel) then "virtual void transfer();\n" else "")
        ++ (if not (null expandCode) then "virtual void expand();\n" else "")
        ++ (if not (null runCode) then "virtual void run();\n" else "")
        ++ (if uncertain then "virtual bool isUncertain() const {return true;}\n" else "")
        ++ ("virtual bool isDiscovery() const {return " ++ (if discovery then "true" else "false") ++ ";}\n")
        ++ "virtual std::string getName() const {return \"" ++ name ++ "\\\\n" ++ descr ++ "\";}\n"
        ++ name ++ "(" ++ paramlLocals locals ++ ") :\n"
        ++ indent "\t" initList ++ " {}\n"
        )
        )
        ++ "};\n"
      src = (tplFunc name "void" "expand()" descr expandCode)
        ++ (tplFunc name "void" "run()" descr runCode)
        ++ (if null commChannel 
              then "" 
              else tplFunc name "void" "transfer()" descr (commChannel ++ "->transfer();\n"))


tplFunc :: String -> String -> String -> String -> String -> String
tplFunc _ _ _ _ "" = ""
tplFunc className typeName name descr code = wrapInFold 
  ("Function `" ++ proto ++ "' (" ++ descr ++ ")") 
  (proto ++ " {\n" ++ (indent "\t" code) ++ "}\n")
  where
    proto = typeName ++ " " ++ className ++ "::" ++ name 


tplTmpProc :: String -> String -> Locals -> String
tplTmpProc name tmpName args = if name == "" 
  then ""
  else if tmpName == "pop"
    then "Proc* " ++ tmpName ++ " = add(new " ++ name ++ "(" ++ arglLocals "this->pop" [] args ++ "));\n"
    else "Proc* " ++ tmpName ++ " = add(new " ++ name ++ "(" ++ arglLocals "pop" [] args ++ "));\n"


-- creates a connect(add(new ....
-- params may be [], in which case no conversion is done for args
tplConnect :: String -> String -> Locals -> Locals -> String
tplConnect name pop params args = if null name
  then ""
  else "connect(add(new " ++ name ++ "(" ++ arglLocals pop params args ++ ")));\n"


-- }}}
