-- {{{ HEADER: File headers
module ToccListVar where

-- This file contains code that lists all variables used in a segment of a
-- program, and in what ways they are used. This information is used in two
-- ways: One, to check for illegal sharing of variables and channels, and two,
-- to check which channels are to be given to which branches in a PAR.
-- Ownership of channels is important to the scheduler, so that it can know
-- where to look for the other end of a channel, when a communication is needed.

import AbsTocc
import PrintTocc
import ErrM
import List
import Maybe

import ToccLocals
import ToccLocalsDecl
import ToccConstExp
import ToccProto
import ToccLib

-- }}}

-- {{{ DATA: VarUse declaration

data VarUse = UseReadOnly | UseReadWrite | UseChannel | UseSend | UseReceive
              deriving(Eq,Ord,Show)

-- }}}

-- {{{ DATA: UsedVar declaration

data UsedVar = UsedVar  { uvName :: String,
                          uvIndex :: LocalIndex,
                          uvUse :: VarUse
                        } deriving (Eq,Ord,Show)

type UsedVars = [UsedVar]


-- }}}

-- {{{ FUNCTIONS: Helper functions.

-- Removes local variables from lists of uses (when popping function calls)
unnestScope :: Locals -> Err UsedVars -> Err UsedVars
unnestScope locals xs = case xs of
  Bad s -> Bad s
  Ok uses -> Ok $ filter (\x -> uvName x `notElem` map locName locals) uses


useChannel :: LowerIdent -> UsedVar
useChannel li = UsedVar (printTree li) IndexNone UseChannel


useReceive :: LowerIdent -> UsedVar
useReceive li = UsedVar (printTree li) IndexNone UseReceive


useSend :: LowerIdent -> UsedVar
useSend li = UsedVar (printTree li) IndexNone UseSend


useReadOnly :: LowerIdent -> UsedVar
useReadOnly li = UsedVar (printTree li) IndexNone UseReadOnly


useReadWrite :: LowerIdent -> UsedVar
useReadWrite li = UsedVar (printTree li) IndexNone UseReadWrite


useIndex :: Locals -> UsedVar -> Exp -> UsedVars
useIndex locals use exp = case evalConstExp locals exp of
  ValueInt i -> [use {uvIndex=IndexInt i}] ++ listvarExp locals exp
  ValueIterator -> [use {uvIndex=IndexIterator}] ++ listvarExp locals exp
  _ -> [use {uvIndex=IndexUnknown}] ++ listvarExp locals exp


toList :: UsedVar -> Err UsedVars
toList x = Ok [x]


uvPrettyName :: UsedVar -> String
uvPrettyName UsedVar {uvName=name, uvIndex=index} = case index of
  IndexNone -> name
  IndexUnknown -> name ++ "[?]"
  IndexIterator -> name ++ "[*current `FOR' iterator*]"
  IndexInt i -> name ++ "[" ++ (show i) ++ "]"


qualFromVarUse :: VarUse -> Qualifier_spec
qualFromVarUse x = case x of
  UseReadOnly -> QualConst
  UseReadWrite -> QualReadWrite
  UseChannel -> QualChannel
  UseSend -> QualChannelOut
  UseReceive -> QualChannelIn


hasKnownIndex :: UsedVar -> Bool
hasKnownIndex x = case uvIndex x of
  IndexInt _ -> True
  _ -> False


localFromUsedVar :: Type_spec -> Array_spec -> LocalValue -> UsedVar -> Local
localFromUsedVar type_spec array_spec value x = case x of
  UsedVar name indexing use -> 
    Local (qualFromVarUse use) type_spec array_spec indexing name value


-- }}}

-- {{{ FUNCTIONS: Combination and validation functions
 
-- Two indices in sequence represent (potentially) the same variable if either
-- index is unknown or none, or if the indices are equal. 
getCommonIndex :: UsedVar -> UsedVar -> Maybe UsedVar
getCommonIndex x y = if uvName x /= uvName y
  then Nothing
  else case (uvIndex x, uvIndex y) of
    (IndexUnknown, _) -> Just x {uvIndex=IndexUnknown}
    (_, IndexUnknown) -> Just x {uvIndex=IndexUnknown}
    (IndexNone, _) -> Just x {uvIndex=IndexNone}
    (_, IndexNone) -> Just x {uvIndex=IndexNone}
    (ix, iy) -> if ix == iy
      then Just x
      else Nothing


-- Base function for determining whether two single uses are allowed in parallel
combineUsesPar :: UsedVar -> UsedVar -> Maybe (Err UsedVar)
combineUsesPar x y = case getCommonIndex x y of
  Nothing -> Nothing
  -- If a variable is indexed with just the current iterator, then it will never
  -- conflict with itself. 
  Just n -> Just $ if uvIndex n == IndexIterator
    then Ok n
    else case (uvUse x, uvUse y) of
      -- Variables: RO+RO=Ok, RW+x=Error
      (UseReadOnly, UseReadOnly) -> Ok n
      (UseReadOnly, UseReadWrite) -> Bad $ "Read/write violation for variable `" ++ uvPrettyName n ++ "'\n"
      (UseReadWrite, UseReadWrite) -> Bad $ "Write/write violation for variable `" ++ uvPrettyName n ++ "'\n"
      -- Channels: Send+Recv=Ok, Send+Send=Error, Recv+Recv=Error. Others=Error
      (UseSend, UseReceive) -> Ok $ n {uvUse=UseChannel}
      (UseReceive, UseSend) -> Ok $ n {uvUse=UseChannel}
      (UseSend, UseSend) -> Bad $ "Send/send violation for channel `" ++ uvPrettyName n ++ "'\n"
      (UseReceive, UseReceive) -> Bad $ "Receive/receive violation for channel `" ++ uvPrettyName n ++ "'\n"
      _ -> Bad $ "Use violation for channel `" ++ uvPrettyName n ++ "'\n"


-- Base function for combining two single uses in sequence.
combineUsesSeq :: UsedVar -> UsedVar -> Maybe (Err UsedVar)
combineUsesSeq x y =  case getCommonIndex x y of
  Nothing -> Nothing
  Just n -> Just $ Ok $ case (uvUse x, uvUse y) of
    -- Variables: RO+RO=RO, RW+x=RW.
    (UseReadOnly, UseReadOnly) -> n
    (UseReadOnly, UseReadWrite) -> n {uvUse=UseReadWrite}
    (UseReadWrite, UseReadWrite) -> n
    -- Channels: Send+Send=Send, Recv+Recv=Recv. Others=Channel (send+recv)
    (UseSend, UseSend) -> n
    (UseReceive, UseReceive) -> n
    _ -> n {uvUse=UseChannel}
    

-- Merges a single use into a list. Should be called with second parameter [].
mergeSingleUseList :: (UsedVar -> UsedVar -> Maybe (Err UsedVar)) -> UsedVars -> UsedVars -> UsedVar -> Err UsedVars
mergeSingleUseList combFun xs [] new = Ok $ new:xs
mergeSingleUseList combFun xs (y:ys) new = case combFun y new of
  Nothing -> mergeSingleUseList combFun (y:xs) ys new
  Just (Ok v) -> mergeSingleUseList combFun xs ys v
  Just (Bad s) -> Bad s


-- Merges two lists of uses.
mergeTwoUseLists :: (UsedVar -> UsedVar -> Maybe (Err UsedVar)) -> UsedVars -> UsedVars -> Err UsedVars
mergeTwoUseLists combFun xs [] = Ok xs
mergeTwoUseLists combFun xs (y:ys) = case mergeSingleUseList combFun [] xs y of 
  Ok u1 -> mergeTwoUseLists combFun u1 ys
  Bad s1 -> case mergeTwoUseLists combFun xs ys of
    Ok _ -> Bad s1
    Bad s2 -> Bad $ s1++s2


-- Merges a list of lists of uses. This function could be a lot simpler if not
-- for wanting to list all errors encountered.
mergeFoldLists :: (UsedVar -> UsedVar -> Maybe (Err UsedVar)) -> Err UsedVars -> [Err UsedVars] -> Err UsedVars
mergeFoldLists combFun ex [] = ex
mergeFoldLists combFun ex (ey:eys) = case ex of
  Bad s -> error "mergeFoldLists init should not be error"
  Ok x -> case ey of 
    Bad s -> case mergeFoldLists combFun ex eys of
      Ok _ -> Bad s
      Bad s2 -> Bad $ s++s2
    Ok y -> case mergeTwoUseLists combFun x y of
      Ok z -> mergeFoldLists combFun (Ok z) eys
      Bad s -> case mergeFoldLists combFun ex eys of
        Ok _ -> Bad s
        Bad s2 -> Bad $ s++s2


-- Merges lists of uses in parallel
catUsesPar :: [Err UsedVars] -> Err UsedVars
catUsesPar useLists = mergeFoldLists combineUsesPar (Ok []) useLists


-- Merges lists of uses in sequence
catUses :: [Err UsedVars] -> Err UsedVars
catUses useLists = mergeFoldLists combineUsesSeq (Ok []) useLists


-- }}}

-- {{{ FUNCTIONS: Splitting locals between PAR branches.

-- Determine which variables and channels go to which branch in a PAR. Returns
-- (if on no errors) a set of (param,arg), where param is the local as it exist
-- in before the PAR, and the arg is how it is used in the specified PAR branch.
-- needLocal takes a single local and returns a list of locals that are used,
-- and that matches the given local.
needLocal :: Program_spec -> Locals -> Proc_spec -> Local -> Err (Maybe (Local, Local))
needLocal ps locals proc x = case listvarProc_spec ps locals proc of
  Bad s -> Bad s
  Ok usedVars -> case find ((==locName x).uvName) usedVars of
    Nothing -> Ok Nothing
    Just y -> if locQualifier x == QualByVal
      then Ok $ Just ((localFromUsedVar (locType x) (locArray x) ValueUnknown y){locIndex=locIndex x,locQualifier=QualByVal}, x)
      else Ok $ Just ((localFromUsedVar (locType x) (locArray x) ValueUnknown y){locIndex=locIndex x}, x)
      

needLocals :: Program_spec -> Locals -> Proc_spec -> Err (Locals, Locals)
needLocals ps locals proc = case mapM (needLocal ps locals proc) locals of
  Bad s -> Bad s
  Ok xs -> Ok $ unzip $ catMaybes xs


-- }}}

-- {{{ FUNCTIONS: SEQ, PAR

listvarSeq :: Program_spec -> Locals -> Block_spec -> Err UsedVars
listvarSeq ps locals block_spec = case block_spec of
  Block variable_specs proc_specs -> 
    unnestScope newLocals $ listvarProc_specs ps (nestScope locals newLocals) proc_specs
    where 
      newLocals = localsVariable_specs locals variable_specs


parProc_specs :: Program_spec -> Locals -> [[Proc_spec]] -> Err UsedVars
parProc_specs _ _ [] = Ok []
parProc_specs ps locals (xs:xss) = 
  catUsesPar [listvarProc_specs ps locals xs, parProc_specs ps locals xss]


listvarPar :: Program_spec -> Locals -> Block_spec -> Err UsedVars
listvarPar ps locals block_spec = case block_spec of
  Block variable_specs proc_specs -> 
    unnestScope newLocals $ parProc_specs ps (nestScope locals newLocals) $ partitionProcs proc_specs
    where 
      newLocals = localsVariable_specs locals variable_specs


listvarOrder :: Program_spec -> Locals -> Order_spec -> Block_spec -> Err UsedVars
listvarOrder ps locals order_spec block_spec = case order_spec of
  OrderSeq -> listvarSeq ps locals block_spec
  OrderPar -> listvarPar ps locals block_spec
  

-- }}}

-- {{{ FUNCTIONS: Replication: SEQ FOR, PAR FOR, ALT FOR

-- FOR: If both bounds can be found at compile-time, All statements are
-- evaluated with the correct iterator, and the result is tested for violations.
-- If the bounds cannot be found (are either known at compile time, or entirely
-- unknown), then this restricts the uses of channel and variable indexes in the
-- branch. If the index is unpredictable, then two statements are necessary to
-- discover all violations (because array sharing is then entirely illegal).
listvarRepl :: ([Err UsedVars] -> Err UsedVars) -> (Program_spec -> Locals -> a -> Err UsedVars) -> Program_spec -> Locals -> LowerIdent -> Exp -> Exp -> a -> Err UsedVars
listvarRepl mergeFun blockFun ps locals li exp0 exp block = catUses $
  [Ok $ listvarExp locals exp0, Ok $ listvarExp locals exp, mergeFun rep]
  where
    rep = case (evalConstExp locals exp0, evalConstExp locals exp) of
      (ValueInt lbi, ValueInt ubi) -> let
        scope i = nestScope (iterToKnown locals) [localsIterVar li (ValueInt i)]
        in map (\i -> blockFun ps (scope i) block) [lbi..ubi] 
      _ -> let
        scope = nestScope (iterToKnown locals) [localsIterVar li ValueIterator]
        in take 2 $ repeat $ blockFun ps scope block


listvarOrderFor :: Program_spec -> Locals -> Order_spec -> LowerIdent -> Exp -> Exp -> SingleStBlock_spec -> Err UsedVars
listvarOrderFor ps locals order li exp0 exp ssblock = case order of
  OrderSeq -> listvarRepl catUses listvarSingleStBlock_spec ps locals li exp0 exp ssblock
  OrderPar -> listvarRepl catUsesPar listvarSingleStBlock_spec ps locals li exp0 exp ssblock


-- }}}

-- {{{ FUNCTIONS: ALT

listvarAlt_spec :: Program_spec -> Locals -> Alt_spec -> Err UsedVars
listvarAlt_spec ps locals x = case x of
  Alt altalt_specs  -> listvarAltAlt_specs ps locals altalt_specs
  AltRepl lowerident exp0 exp altalt_specs  ->
    listvarRepl catUses listvarAltAlt_specs ps locals lowerident exp0 exp altalt_specs


listvarAltComm_spec :: AltComm_spec -> Err UsedVars
listvarAltComm_spec x = case x of
  AltCommRecv lowerident0 lowerident  -> catUses [toList $ useReceive lowerident0, toList $ useReadOnly lowerident]


listvarAltAlt_specs :: Program_spec -> Locals -> [AltAlt_spec] -> Err UsedVars
listvarAltAlt_specs ps locals xs = catUses $ map (listvarAltAlt_spec ps locals) xs


listvarAltAlt_spec :: Program_spec -> Locals -> AltAlt_spec -> Err UsedVars
listvarAltAlt_spec ps locals x = case x of
  AltAltNested alt_spec  -> listvarAlt_spec ps locals alt_spec
  AltAltGuard exp altcomm_spec ssblock_spec  -> 
    catUses [Ok $ listvarExp locals exp, 
      listvarAltComm_spec altcomm_spec, listvarSingleStBlock_spec ps locals ssblock_spec]
  AltAltSkip exp  -> Ok $ listvarExp locals exp
  AltAlt altcomm_spec ssblock_spec  -> 
    catUses [listvarAltComm_spec altcomm_spec,
      listvarSingleStBlock_spec ps locals ssblock_spec]


-- }}}

-- {{{ FUNCTIONS: Assign/AssignArray

listvarAssign' :: Program_spec -> Locals -> UsedVar -> Exp -> Err UsedVars
listvarAssign' ps locals var exp = case find (\x -> locName x == uvName var) locals of
  Nothing -> Bad $ "Variable `" ++ uvName var ++ "' not found.\n"
  Just loc -> if locQualifier loc /= QualReadWrite 
    then Bad $ "Assigning to read/only variable `" ++ uvName var ++ "'\n"
    else catUses [Ok $ listvarExp locals exp, toList $ var]


listvarAssign :: Program_spec -> Locals -> LowerIdent -> Exp -> Err UsedVars
listvarAssign ps locals li exp = listvarAssign' ps locals (useReadWrite li) exp


listvarAssignArr :: Program_spec -> Locals -> LowerIdent -> Exp -> Exp -> Err UsedVars
listvarAssignArr ps locals li indexExp exp = catUses $ [Ok $ listvarExp locals indexExp, 
  listvarAssign' ps locals ((useReadWrite li) {uvIndex=i}) exp]
  where
    i = case evalConstExp locals indexExp of
      ValueInt i -> IndexInt i
      _ -> IndexUnknown
 

-- }}}

-- {{{ FUNCTIONS: Other parts of syntax tree

listvarProgram_spec :: Program_spec -> Err UsedVars
listvarProgram_spec x = case x of
  Program global_specs  -> listvarGlobal_specs x global_specs


listvarGlobal_specs :: Program_spec -> [Global_spec] -> Err UsedVars
listvarGlobal_specs ps xs = catUses $ map (listvarGlobal_spec ps) xs


listvarGlobal_spec :: Program_spec -> Global_spec -> Err UsedVars
listvarGlobal_spec ps x = case x of
  GlobalProcess capident param_specs singlestblock_spec  ->
    case unnestScope newLocals $ listvarSingleStBlock_spec ps newLocals singlestblock_spec of
      Ok uses -> Ok uses
      Bad s -> Bad $ "In process `" ++ (printTree capident) ++ "':\n" ++ indent " " s
      where 
        newLocals = byValToConst $ localsParam_specs param_specs
  GlobalTypedef _ _ _  -> Ok []
  GlobalInclude _  -> Ok []
  GlobalSInclude _  -> Ok []


listvarSingleStBlock_spec :: Program_spec -> Locals -> SingleStBlock_spec -> Err UsedVars
listvarSingleStBlock_spec ps locals x = case x of
  SingleStBlock variable_specs proc_specs  -> 
    unnestScope newLocals $ listvarProc_specs ps (nestScope locals newLocals) proc_specs
    where 
      newLocals = localsVariable_specs locals variable_specs


listvarComm_spec :: Comm_spec -> Err UsedVars
listvarComm_spec x = case x of
  CommSend lowerident0 lowerident  -> catUses [toList $ useSend lowerident0, toList $ useReadOnly lowerident]
  CommRecv lowerident0 lowerident  -> catUses [toList $ useReceive lowerident0, toList $ useReadWrite lowerident]


listvarArg_spec :: Program_spec -> Locals -> String -> (Int, Arg_spec) -> Err UsedVars
listvarArg_spec ps locals proc (i,x) = case x of 
  Arg exp -> case lookupProto ps proc i of
    Bad s -> Bad s
    Ok local -> case locQualifier local of
      QualByVal -> Ok $ listvarExp locals exp
      q -> case exp of
        ExpVar lowerident -> toList $ useFun lowerident
        ExpIndex lowerident exp -> Ok $ useIndex locals (useFun lowerident) exp
        _ -> Bad $ "Cannot pass expression by reference, parameter #" ++ (show i) ++ " of process `" ++ proc ++ "'\n"
        where 
          useFun = case q of
            QualChannel -> useChannel
            QualChannelIn -> useReceive
            QualChannelOut -> useSend
            QualReadWrite -> useReadWrite
            QualConst -> useReadOnly


listvarArg_specs :: Program_spec -> Locals -> String -> [Arg_spec]-> Err UsedVars
listvarArg_specs ps locals proc xs = catUses $ map (listvarArg_spec ps locals proc) $ zip [1..] xs


-- }}}

-- {{{ FUNCTIONS: Exp

listvarExp :: Locals -> Exp -> UsedVars
listvarExp locals x = case x of
  ExpLor exp0 exp      -> listvarExp locals exp0 ++ listvarExp locals exp
  ExpLand exp0 exp     -> listvarExp locals exp0 ++ listvarExp locals exp
  ExpBitor exp0 exp    -> listvarExp locals exp0 ++ listvarExp locals exp
  ExpBitexor exp0 exp  -> listvarExp locals exp0 ++ listvarExp locals exp
  ExpBitand exp0 exp   -> listvarExp locals exp0 ++ listvarExp locals exp
  Expeq exp0 exp       -> listvarExp locals exp0 ++ listvarExp locals exp
  Expneq exp0 exp      -> listvarExp locals exp0 ++ listvarExp locals exp
  Explthen exp0 exp    -> listvarExp locals exp0 ++ listvarExp locals exp
  Expgrthen exp0 exp   -> listvarExp locals exp0 ++ listvarExp locals exp
  Exple exp0 exp       -> listvarExp locals exp0 ++ listvarExp locals exp
  Expge exp0 exp       -> listvarExp locals exp0 ++ listvarExp locals exp
  ExpLeft exp0 exp     -> listvarExp locals exp0 ++ listvarExp locals exp
  ExpRight exp0 exp    -> listvarExp locals exp0 ++ listvarExp locals exp
  ExpPlus exp0 exp     -> listvarExp locals exp0 ++ listvarExp locals exp
  ExpMinus exp0 exp    -> listvarExp locals exp0 ++ listvarExp locals exp
  ExpTimes exp0 exp    -> listvarExp locals exp0 ++ listvarExp locals exp
  ExpDiv exp0 exp      -> listvarExp locals exp0 ++ listvarExp locals exp
  ExpMod exp0 exp      -> listvarExp locals exp0 ++ listvarExp locals exp
  ExpNegative exp      -> listvarExp locals exp
  ExpPositive exp      -> listvarExp locals exp
  ExpIndex lowerident exp  -> useIndex locals (useReadOnly lowerident) exp
  ExpFunc lowerident exps  -> concat $ map (listvarExp locals) exps
  ExpConst constval_spec  -> []
  ExpVar lowerident  -> [useReadOnly lowerident]


-- }}}

-- {{{ FUNCTIONS: Proc_spec/specs, distributing function.

-- Short hand for any element that has an exp and a block.
listvarExpBlock :: Program_spec -> Locals -> Exp -> SingleStBlock_spec -> Err UsedVars
listvarExpBlock ps locals exp ssblock = catUses $ 
  [Ok $ listvarExp locals exp, listvarSingleStBlock_spec ps locals ssblock]


listvarProc_specs :: Program_spec -> Locals -> [Proc_spec] -> Err UsedVars
listvarProc_specs ps locals xs = catUses $ map (listvarProc_spec ps locals) xs


listvarProc_spec :: Program_spec -> Locals -> Proc_spec -> Err UsedVars
listvarProc_spec ps locals x = case x of
  ProcAssign lowerident assignment_op exp  -> listvarAssign ps locals lowerident exp
  ProcAssignArr lowerident exp0 assignment_op exp  -> 
    listvarAssignArr ps locals lowerident exp0 exp
  ProcFor order_spec lowerident exp0 exp singlestblock_spec  -> 
    listvarOrderFor ps locals order_spec lowerident exp0 exp singlestblock_spec
  ProcWhile exp singlestblock_spec  -> listvarExpBlock ps locals exp singlestblock_spec
  ProcAlt alt_spec  -> listvarAlt_spec ps locals alt_spec
  ProcIf exp singlestblock_spec  -> listvarExpBlock ps locals exp singlestblock_spec
  ProcElse singlestblock_spec  -> listvarSingleStBlock_spec ps locals singlestblock_spec
  ProcElseIf exp singlestblock_spec  -> listvarExpBlock ps locals exp singlestblock_spec
  ProcOrder order_spec block_spec  -> listvarOrder ps locals order_spec block_spec
  ProcTime exp singlestblock_spec  -> listvarExpBlock ps locals exp singlestblock_spec
  ProcDelay exp  -> Ok $ listvarExp locals exp
  ProcExp exp  -> Ok $ listvarExp locals exp
  ProcPrint exps  -> Ok $ concat $ map (listvarExp locals) exps
  ProcSkip  -> Ok []
  ProcStop  -> Ok []
  ProcComm comm_spec  -> listvarComm_spec comm_spec
  ProcCall capident arg_specs  -> listvarArg_specs ps locals (printTree capident) arg_specs


-- }}}

-- {{{ FUNCTIONS: Entry function checking for shared vars or doubly used channels.

varchkProgram_spec :: Err Program_spec -> Maybe String
varchkProgram_spec x = case x of 
    Ok ps -> case listvarProgram_spec ps of
      Ok _ -> Nothing
      Bad s -> Just s
    Bad s -> Just s


-- }}}

