-- {{{ HEADER: File headers
module ToccLocals where

-- This file contans code that extracts Local lists from various program
-- declarations. A Local is a generic variable/parameter. General code for
-- nesting scopes etcetera is also in this file. The definition of Local itself
-- is in ToccLocalsDecl, to avoid a circular dependency with ToccConstExp.

import AbsTocc
import PrintTocc
import ErrM
import List

import ToccLocalsDecl
import ToccConstExp

-- }}}

-- {{{ FUNCTIONS: Helper functions
 
-- Remove locals from previous scope that conflicts with new.
filterScope :: Locals -> Locals -> Locals
filterScope old new = filter (\x -> locName x `notElem` map locName new) old


-- Combine new locals with remaining old locals in a scope.
nestScope :: Locals -> Locals -> Locals
nestScope old new = filterScope old new ++ new


-- Remap all byVal Qualifiers to Const. Byval is only used in the opening of a
-- named process. After this, it is passed internally by reference.
byValToConst :: Locals -> Locals
byValToConst [] = []
byValToConst (x:xs) = (if locQualifier x==QualByVal then x {locQualifier=QualConst} else x) : byValToConst xs


-- Create an iterator-style local variable.
localsIterVar :: LowerIdent -> LocalValue -> Local
localsIterVar li v = 
  defLocal {locQualifier=QualByVal, locType=TypeInt, locName=printTree li, locValue=v} 


-- Remove previous current iterator flag from a scope.
iterToKnown :: Locals -> Locals
iterToKnown [] = []
iterToKnown (x:xs) = (if locValue x==ValueIterator then x {locValue=ValueKnownExpandTime} else x) : iterToKnown xs


-- When adding token position information, these functions make more sense..
-- (they will be ..=LowerIdent ((0,0), s))
lowerIdent :: String -> LowerIdent
lowerIdent s = LowerIdent s


capIdent :: String -> CapIdent
capIdent s = CapIdent s


upperIdent :: String -> UpperIdent
upperIdent s = UpperIdent s


typeCustom :: String -> Type_spec
typeCustom s = TypeCustom $ upperIdent s


-- }}}

-- {{{ FUNCTIONS: Create Locals from declarations/variables

localsParam_specs :: [Param_spec] -> Locals
localsParam_specs xs = map localsParam_spec xs
 

localsParam_spec :: Param_spec -> Local
localsParam_spec x = case x of
  Param decl_spec  -> localDecl_spec decl_spec


localInit_spec :: Locals -> Init_spec -> Local -> Local
localInit_spec locals init x = if locQualifier x == QualConst && locType x == TypeInt
  then case init of
    InitExp exp  -> x {locValue=evalConstExp locals exp}
    InitNone  -> x
  else x


-- To create a locals from variable specs one need the current Locals scope in
-- order to get the value right. If the compile-time value is not interesting,
-- an empty locals can be given.
localsVariable_specs :: Locals -> [Variable_spec] -> Locals
localsVariable_specs locals xs = map (localVariable_spec locals) xs
 

localVariable_spec :: Locals -> Variable_spec -> Local
localVariable_spec locals x = case x of
  Var decl_spec init_spec  -> localInit_spec locals init_spec $ localDecl_spec decl_spec


localsDecl_specs :: [Decl_spec] -> Locals
localsDecl_specs xs = map localDecl_spec xs
 

localDecl_spec :: Decl_spec -> Local
localDecl_spec x = case x of
  Decl qualifier_spec type_spec array_spec lowerident  -> 
    defLocal {locQualifier=qualifier_spec, locType=type_spec, locArray=array_spec, locName=printTree lowerident}


-- }}}


