-- {{{ HEADER: File headers
module ToccOpt (parseOpts, Action(..)) where

-- This file contains code that parses command line arguments. It uses the
-- GetOpt library.

import System.Console.GetOpt
import System.IO
import Data.Maybe

import System.Exit
import Data.List

-- }}}

-- {{{ DATA: Action

data Action = ActionCheck      { checkVerbose    :: Bool,
                                 checkFiles      :: [String]
                               } 
            | ActionCompile    { compileVerbose  :: Bool,
                                 compileMakefile   :: String,
                                 compileOutput    :: String,
                                 compileFiles    :: [String]
                               } deriving (Ord,Eq,Show)

-- }}}

-- {{{ DATA: Flag

data Flag
  = Check          -- V
  | Verbose         -- v
  | Output String      -- o
  | Makefile String    -- m
  | Help          -- h
  | Version        -- h
  deriving (Ord,Eq,Show)

-- }}}

-- {{{ OPTIONS: List of options and help text

flags =
  [Option ['c'] ["check"]    (NoArg Check)
    "Run syntax check only. Used with -v prints clean AST."
  ,Option ['v'] ["verbose"]  (NoArg Verbose)
    "Be verbose."
  ,Option ['o'] ["output"]  (ReqArg Output "BASENAME")
    "Select output file name base. Files are named BASENAME.cpp/.hpp."
  ,Option ['m'] ["makefile"]  (OptArg optMakefile "FILE")
    "Create Makefile as well. Default Makefile name is 'Makefile'."
  ,Option ['h','?'] ["help"]  (NoArg Help)
    "Show this help message."
  ,Option ['V'] ["version"]  (NoArg Version)
    "Display version information."
  ]

-- Type of optional options are Maybe String. These provides a default.
optMakefile :: Maybe String -> Flag
optMakefile = Makefile . fromMaybe "Makefile"


-- }}}

-- {{{ FUNCTIONS: parseOpts, getMakefile, getOutput

-- Get flag parameters
getMakefile :: [Flag] -> String
getMakefile [] = ""
getMakefile (x:xs) = 
  case x of 
    Makefile s -> s
    _ -> getMakefile xs

getOutput :: [Flag] -> String
getOutput [] = ""
getOutput (x:xs) = case x of
  Output s -> s
  _ -> getOutput xs

-- Main parsing function
parseOpts :: [String] -> IO Action
parseOpts argv = case getOpt Permute flags argv of
  (args,files,[]) -> do
    let help = Help `elem` args
    let version = Version `elem` args
    let check = Check `elem` args
    let verbose = Verbose `elem` args
    let makefile = getMakefile args
    let output = getOutput args
    case (help, version, files, check, makefile, output) of
      (True,_,_,_,_,_) -> do
        hPutStrLn stderr (usageInfo header flags)
        exitWith ExitSuccess
      (_,True,_,_,_,_) -> do
        hPutStrLn stderr "tocc version 0.1, Martin Korsgaard 2007"
        exitWith ExitSuccess
      (_,_,[],_,_,_) -> do
        hPutStrLn stderr "tocc: No input file."
        exitWith (ExitFailure 1)
      (_,_,_,True,"","") -> do
        return (ActionCheck verbose files)
      (_,_,_,True,"",_) -> do
        hPutStrLn stderr "tocc: Cannot specify output file when syntax checking only."
        exitWith (ExitFailure 1)
      (_,_,_,True,_,_) -> do
        hPutStrLn stderr "tocc: Cannot specify makefile file when syntax checking only."
        exitWith (ExitFailure 1)
      _ -> do
        return (ActionCompile verbose makefile output files)

  (_,_,errs) -> do
    hPutStrLn stderr (concat errs ++ usageInfo header flags)
    exitWith (ExitFailure 1)

  where header = "Usage: tocc [-chmvV] [-o base] file"

-- }}}

