#ifndef __INCLUDE_CHANNEL_HPP__
#define __INCLUDE_CHANNEL_HPP__

#include "Proc.hpp"

// Forward declarations.
template <class T> class Channel;
class ChannelBase;

// SendChannel is a Channel class restricted to sending.
template <class T>
class SendChannel {
	friend class Channel<T>;
	private:
		Channel<T>* ch;

		SendChannel(Channel<T>* ch) : ch(ch) {}

	public:
		// Return pointer to channel.
		Channel<T>* getChannel() const {
			return ch;
		}

		// Return pointer to type-independent channel base.
		ChannelBase* getChannelBase() {
			return ch->getChannelBase();
		}

		void setSenderOwner(Proc* t) {
			ch->setSenderOwner(t);
		}

		void setOwner(Proc* t) {
			ch->setSenderOwner(t);
		}

		Proc* getSenderOwner() const {
			return ch->getSenderOwner();
		}

		Proc* getOwner() const {
			return ch->getSenderOwner();
		}

		void transfer() {
			ch->transfer();
		}

		void send(const T* ptr) {
			ch->send(ptr);
		}

};

template <class T>
class ReceiveChannel {
	friend class Channel<T>;
	private:
		Channel<T>* ch;

		ReceiveChannel(Channel<T>* ch) : ch(ch) {}

	public:
		// Return pointer to channel.
		Channel<T>* getChannel() const {
			return ch;
		}

		// Return pointer to type-independent channel base.
		ChannelBase* getChannelBase() {
			return ch->getChannelBase();
		}

		void setReceiverOwner(Proc* t) {
			ch->setReceiverOwner(t);
		}

		void setOwner(Proc* t) {
			ch->setReceiverOwner(t);
		}

		Proc* getReceiverOwner() const {
			return ch->getReceiverOwner();
		}

		Proc* getOwner() const {
			return ch->getReceiverOwner();
		}

		void transfer() {
			ch->transfer();
		}

		void receive(T* ptr) {
			ch->receive(ptr);
		}
};

// Type-independent channel base
class ChannelBase {
	private:
		// Which procs to expand if sender or receiver is missing?
		Proc* senderOwner;
		Proc* receiverOwner;

	protected:
		bool senderReady;
		bool receiverReady;
	public:
		ChannelBase() : 
				senderOwner(NULL), receiverOwner(NULL), senderReady(false),
				receiverReady(false) {}

		bool isReceiverReady() const {
			return receiverReady;
		}

		bool isSenderReady() const {
			return senderReady;
		}

		ChannelBase* getChannelBase() {
			return this;
		}

		// set ownership
		void setOwner(Proc* t) {
			senderOwner = t;
			receiverOwner = t;
		}

		// set sender ownership
		void setSenderOwner(Proc *t) {
			senderOwner = t;
		}

		// set receiver ownership
		void setReceiverOwner(Proc *t) {
			receiverOwner = t;
		}

		// Get sender ownership
		Proc* getSenderOwner() const {
			return senderOwner;
		}

		// Get receiver ownership
		Proc* getReceiverOwner() const {
			return receiverOwner;
		}

};

template<class T>
class Channel : public ChannelBase {
	private:
		const T* sendPtr;
		T* receivePtr;
		// receiveChannel and sendChannels
		ReceiveChannel<T> receiveChannel;
		SendChannel<T> sendChannel;


	public:
		Channel() : 
			ChannelBase(), receiveChannel(this), sendChannel(this) {}

		// transfer() copies the actual value being transferred, and completes
		// the transaction.
		void transfer() {
			assert(senderReady && receiverReady);
			*receivePtr = *sendPtr;
			senderReady = false;
			receiverReady = false;
		}

		// send() registers a sender, but does not do any sending.
		void send(const T* ptr) {
			assert(ptr);
			assert(!senderReady);

			this->sendPtr = ptr;
			senderReady = true;
		}

		// receive() registers a receiver, but does not do any receiving.
		void receive(T* ptr) {
			assert(ptr);
			assert(!receiverReady);

			this->receivePtr = ptr;
			receiverReady = true;
		}

		// Return a receive-only channel pointer.
		ReceiveChannel<T>* getReceiveChannel() {
			return &receiveChannel;
		}

		// Return a send-only channel pointer.
		SendChannel<T>* getSendChannel() {
			return &sendChannel;
		}
};

// The ChannelVector must be a separate class, to avoid memory leak problems
// caused by creating SendChannel and ReceiveChannel vectors on the fly.
// SendChannel and ReceiveChannel vectors are standard vectors.
template <class T>
class ChannelVector {
	private:
		std::vector<Channel<T>*> chv;
		std::vector<ReceiveChannel<T>*> rchv;
		std::vector<SendChannel<T>*> schv;

		// Direct copy is illegal:
		ChannelVector(const ChannelVector<T>& rhs); 

	public:
		~ChannelVector() {
			for (size_t i = 0; i < chv.size(); ++i) {
				delete chv[i];
			}
		}

		ChannelVector() {
			
		}

		ChannelVector(int n) : chv(n), rchv(n), schv(n) {
			for (size_t i = 0; i < chv.size(); ++i) {
				chv[i] = new Channel<T>;
				rchv[i] = chv[i]->getReceiveChannel();
				schv[i] = chv[i]->getSendChannel();
			}
		}

		std::vector<ReceiveChannel<T>*>* getReceiveChannelVector() {
			return &rchv;
		}

		std::vector<SendChannel<T>*>* getSendChannelVector() {
			return &schv;
		}

		Channel<T>* operator [] (size_t index) {
			return chv[index];
		}
};


#endif //#ifndef __INCLUDE_CHANNEL_HPP__

