#ifndef __INCLUDE_COMMPROC_HPP__
#define __INCLUDE_COMMPROC_HPP__

#include "Proc.hpp"
#include "Channel.hpp"

class CommProc : public Proc {
	public:
		enum Side {SENDER, RECEIVER};
	private:
		CommProc* matching;
		ChannelBase* chb;
		Side side;
	public:
		// One constructor for sends.
		template <class T>
		CommProc(Proc* pop, SendChannel<T>* sch, const T* var) :
				Proc(pop), matching(NULL), chb(sch->getChannelBase()),
				side(SENDER) {
			sch->send(var);
		}

		// One constructor for receives.
		template <class T>
		CommProc(Proc* pop, ReceiveChannel<T>* rch, T* var) :
				Proc(pop), matching(NULL), chb(rch->getChannelBase()),
				side(RECEIVER) {
			rch->receive(var);
		}

		Side getSide() const {
			return side;
		}

		// Returns proc that owns the other side of the channel.
		Proc* getOtherOwner() const {
			if (side == SENDER)
				return chb->getReceiverOwner();
			else
				return chb->getSenderOwner();
		}

		// Returns whether or not the transfer has been completed.
		bool isCompleted() const {
			return matching != NULL;
		}

		// Overloaded by each subclass, because the CommProc class is
		// channel-type independent.
		virtual void transfer() = 0;

		// Return whether or not the other side is ready to transfer.
		bool isOtherReady() const {
			if (side == SENDER)
				return chb->isReceiverReady();
			else
				return chb->isSenderReady();
		}

		// Check if it is possible to complete a transfer (send/receive) on this
		// channel. If the transfer can be made, this function completes the
		// transfer and returns true. If the transfer has already been
		// completed, it also returns true. If the transfer cannot be made, it
		// returns false.
		bool doTransfer() {
			if (matching != NULL) {
				return true;
			}
			else if (!chb->isSenderReady() || !chb->isReceiverReady()) {
				return false;
			}
			else {
				CommProc* other = dynamic_cast<CommProc*>(getOtherOwner());
				assert(other);
				assert(other->matching == NULL);

				transfer();

				this->matching = other;
				other->matching = this;

				return true;
			}
		}

		// Returns extra graph data for dot graphs.
		virtual std::string getDotExtra() const {
			Proc* other = getOtherOwner();

			if (isCompleted()) {
				assert(this == matching->matching);
				return std::string() + getDotDescription() + " -> " +
					this->matching->getDotDescription() + "[dir=both];\n";
			}
			else {
				return std::string() + getDotDescription() + " -> " +
					other->getDotDescription() +
					"[dir=none,style=dotted,label=\"depends on\"];\n";
			}
		}
};

#endif //#ifndef __INCLUDE_COMMPROC_HPP__

