#ifndef __INCLUDE_PROC_HPP__
#define __INCLUDE_PROC_HPP__

#include "lib.hpp"
#include "take.hpp"
#include "Scheduler.hpp"

#include <list>
#include <string>
#include <cassert>
#include <algorithm>

class Proc {
	private:
		static unsigned gProcIndex;
		const unsigned procIndex;
		
		// ProcState is defined in Scheduler.hpp
		ProcState state;

		// Child of proc.
		Proc* child;
		
	protected:
		// pop is the next proc of the calling process, the one that should be
		// run if this chain of procs ends.
		Proc* const pop;
		
		Proc(Proc* pop) : 
			procIndex(gProcIndex++), state(CLOSED), child(NULL), pop(pop) {}

		// connect() creates a dependency link from *this to child.
		virtual Proc* connect(Proc* child) {
			assert(this->child == NULL);
			assert(child);
			this->child = child;
			child->addParent();
			return child;
		}
		
		// Return count of children. More important for ParProc.
		virtual size_t countChildren() const {
			return child != NULL;
		}

		// Adds a new proc to global list of procs. Calls Scheduler::add.
		template <class ProcClass>
		static ProcClass* add(ProcClass* t) {
			Scheduler::Instance().add(t);
			return t;
		}

		// Removes a new proc from global list of procs. Calls
		// Scheduler::remove.
		template <class ProcClass>
		static ProcClass* remove(ProcClass* t) {
			Scheduler::Instance().remove(t);
			return t;
		}

		public:
		virtual ~Proc() {}

		// addParent is used by ParEnd to count how many procs that have
		// joined. It is empty for others.
		virtual void addParent() {

		}

		virtual void removeParent() {

		}

		// The default expand is to connect to pop.
		virtual void expand() {
			connect(pop);
		}

		virtual Proc* getPop() const {
			return pop;
		}

		virtual Proc* getChild() const {
			return child;
		}

		ProcState getState() const {
			return state;
		}

		void setState(ProcState state) {
			this->state = state;
		}

		unsigned getProcIndex() const {
			return procIndex;
		}

		virtual bool isUncertain() const {
			return false;
		}

		template <class T>
		T* take(T* var) {
			doTake(this, var);
			return var;
		}

		// getDot returns a graphwiz dot representation of a proc node and all
		// links to children.
		std::string getDot() const;

		// getDescription returns the text in the graphwiz dot representation.
		// This is unique, so creating a dot link to `getDescription` will
		// create an arrow into this proc node.
		virtual std::string getDotDescription() const;

		// Style is the shape and style of node to use in graphwiz dot.
		// Depends on state, and is overloaded by channelproc.
		virtual std::string getDotStyle() const;

		// Name is the name of type of proc, such as "Main\\nPar" or
		// "Decode_3\\nAlt"
		virtual std::string getName() const;

		// Extra dot commands. Can be redefined by inheritance.
		virtual std::string getDotExtra() const {return "";}
		virtual std::string getDotDescriptionExtra() const {return "";}
};

#endif //#ifndef __INCLUDE_PROC_HPP__

