// {{{ #Includes 
#include "TimerProc.hpp"
#include "TimerEndProc.hpp"
#include "CommProc.hpp"
#include "ControlProc.hpp"
#include "CodeProc.hpp"
#include "ParProc.hpp"
#include "ParEndProc.hpp"

#include "Channel.hpp"
#include "Scheduler.hpp"
#include "lib.hpp"

#include "Main.hpp"
#include "Stop.hpp"

#include <list>
#include <cstdio>
#include <string>
#include <algorithm>

using namespace std;
// }}}

// {{{ void Scheduler::doAddTimer(...) 
void Scheduler::doAddTimer(TimerProc* timer, int startTime) {
	timer->setStartTime(startTime);

	list<TimerProc*>::iterator it = 
		upper_bound(timers.begin(), timers.end(), timer, DeadlineCompare());

	timers.insert(it, timer);

}
// }}}

// {{{ void Scheduler::doRemoveTimer(...) 
void Scheduler::doRemoveTimer(TimerProc* timer) {
	list<TimerProc*>::iterator it = 
		find(timers.begin(), timers.end(), timer);
	assert(it != timers.end());
	timers.erase(it);
}
// }}}

// {{{ void Scheduler::doSerialize(...) 
void Scheduler::doSerialize(Proc* t) {
	if (t->getState() == CLOSED) {
		doExpand(t);
	}

	if (t->getState() == EXPANDED) {
		t->setState(SERIALIZED);

		// Remove this if to keep all procs in serialized queue, even those that
		// need not run.
		if (dynamic_cast<CodeProc*>(t) != NULL) {
			printf("SERIALIZE  %s\n", t->getDotDescription().c_str());
			serialized.push_back(t);
		}
	}
}
// }}}

// {{{ void Scheduler::doExpand(...) 
void Scheduler::doExpand(Proc* t) {
	if (t->getState() == CLOSED) {
		t->expand();
		t->setState(EXPANDED);
	}
}
// }}}

// {{{ Proc* Scheduler::remove(...) 
Proc* Scheduler::remove(Proc* t) {
	assert(t);
	printf("REMOVE     %s\n", t->getDotDescription().c_str());
	procs.remove(t);
	return t;
}
// }}}

// {{{ Proc* Scheduler::add(...) 
Proc* Scheduler::add(Proc* t) {
	assert(t);
	printf("CREATE     %s\n", t->getDotDescription().c_str());
	procs.push_back(t);
	return t;
}
// }}}

// {{{ void Scheduler::print(...) 
void Scheduler::print(bool printSerialized=false) {
	static int graphIndex = 0;

	assert(graphIndex < 100);

	system("mkdir -p dots");
	string filename = "dots/sched" + toStr(graphIndex++) + ".dot";
	printf("PRINT     %s\n", filename.c_str());
	FILE *of = fopen(filename.c_str(), "w");

	if (of == NULL) {
		perror("Error opening output file");
		return;
	}

	fprintf(of, "digraph P {\n");

	for (ProcIt it = procs.begin(); it != procs.end(); ++it) {
		fprintf(of, "%s", (*it)->getDot().c_str());
	}

	if (printSerialized) {
		int i = 1;
		for (deque<Proc*>::iterator it = serialized.begin();
				it != serialized.end() && it+1 != serialized.end() ; ++it, ++i) {
			fprintf(of, "%s -> %s [color=gray,constraint=false,label=\"%d\"];\n", 
					(*it)->getDotDescription().c_str(),
					(*(it+1))->getDotDescription().c_str(), i);
		}
	}

	fprintf(of, "}\n");

	fclose(of);
}
// }}}

// serialize delegates to either of the other serializing functions depending on
// what type of proc is next. It returns when there is a choice between which
// proc is next, or when time horizon is up.
// {{{ void Scheduler::serialize(...) 
void Scheduler::serialize(TimerProc* active, int horizon, SerializeResult* result) {
	assert(active);

	printf("SERIALIZE Active timer: %s\n", active->getDotDescription().c_str());
	// Pick next proc from active timer.
	Proc* next = active->getNext();
	Proc* proc = NULL;
	*result = CONTINUE;

	// If any of the delegated functions return NULL as next proc, return to
	// select.
	while (*result == CONTINUE) {
		TimerProc* timer;
		TimerEndProc* timerEnd;
		ControlProc* ctl;
		CommProc* comm;
		CodeProc* code;
		ParProc* par;
		ParEndProc* parEnd;

		proc = next;
		if ((timer = dynamic_cast<TimerProc*>(proc)) != NULL) 
			next = serializeTimerProc(timer, result);
		else if ((timerEnd = dynamic_cast<TimerEndProc*>(proc)) != NULL) 
			next = serializeTimerEndProc(timerEnd, result);
		else if ((ctl = dynamic_cast<ControlProc*>(proc)) != NULL) 
			next = serializeControlProc(ctl, result);
		else if ((comm = dynamic_cast<CommProc*>(proc)) != NULL) 
			next = serializeCommProc(comm, result);
		else if ((par = dynamic_cast<ParProc*>(proc)) != NULL) 
			next = serializeParProc(par, result);
		else if ((parEnd = dynamic_cast<ParEndProc*>(proc)) != NULL) 
			next = serializeParEndProc(parEnd, result);
		else if ((code = dynamic_cast<CodeProc*>(proc)) != NULL) 
			next = serializeCodeProc(code, &horizon, result);
		else
			assert(!"Proc is of unknown type!");

	}

	// `proc` is the last proc not to expand. Active timer is set to begin
	// expanding at this proc the next time. Note that this is not important,
	// only faster. There may be more than one timer spanning this chain of
	// proc, but only the latest will have its timer updated. The other timers
	// must iterate through the already expanded procs themselves next time.
	active->setNext(proc);

	print();
}
// }}}

// If a timer is encountered, expand it, add it to timers list, and marked as
// SERIALIZED. If it is already serialized, do nothing with it. Next proc to
// be serialized is always the following proc.
// {{{ Proc* Scheduler::serializeTimerProc(...) 
Proc* Scheduler::serializeTimerProc(TimerProc* timer, SerializeResult* result) {
	assert(timer->getState() != EXPANDED);

	if (timer->getState() == CLOSED) {
		doSerialize(timer);
		doAddTimer(timer, predictionTime);
	}
	return timer->getChild();
}
// }}}

// Timer End: When first expanding the timer end, also remove the timer from the
// timers deque. If there are adjacent new timers, these must be discovered as
// well. If the deadline is missed, the discovery time is the prediction time
// (the next tasks will be started directly after the timer finished, even if it
// finished too late). If the deadline as met, the discovery time is the
// deadline, because following tasks are not allowed to start too early. Return
// RESELECT if deadline met, return OVERFLOW if not.
// {{{ Proc* Scheduler::serializeTimerEndProc(...) 
Proc* Scheduler::serializeTimerEndProc(TimerEndProc* timerEnd, SerializeResult* result) {
	if (timerEnd->getState() == CLOSED || timerEnd->getState() == EXPANDED) {
		TimerProc* timer = timerEnd->getTimerProc();
		int deadline = timer->getStartTime() + timer->getTime();
		doSerialize(timerEnd);
		doRemoveTimer(timer);
		
		if (predictionTime > deadline) {
			discover(timerEnd, predictionTime);
			*result = OVERFLOW;
		}
		else {
			discover(timerEnd, deadline);
			*result = RESELECT;
		}

		return NULL;
	}
	return timerEnd->getChild();
}
// }}}

// For a control proc, just expand it, and move `next`. Do not add to
// serialized, as there is no code to be run anyway. Next proc to be
// serialized is always the following proc. If the control proc is uncertain
// (does not know its children), set global Scheduler::uncertain flag, and
// return to select, which will then run everything up to this point.
// {{{ Proc* Scheduler::serializeControlProc(...) 
Proc* Scheduler::serializeControlProc(ControlProc* ctl, SerializeResult* result) {
	if (ctl->isUncertain()) {
		*result = UNCERTAIN;
		return NULL;
	}

	doSerialize(ctl);
	return ctl->getChild();
}
// }}}

// For a comm. proc. If the communication is ready, then we expand it, and
// return to select. It is necessary to return to select, because at this point
// we have a choice for which proc to continue serializing. If the other side
// of communication is not ready, we do not have a choice, so we must continue
// expanding at that side.
// {{{ Proc* Scheduler::serializeCommProc(...) 
Proc* Scheduler::serializeCommProc(CommProc* comm, SerializeResult* result) {
	if (comm->getState() != CLOSED) {
		return comm->getChild();
	}
	else if (comm->doTransfer()) {
		doSerialize(comm);
		*result = RESELECT;
		return NULL;
	}
	else {
		return comm->getOtherOwner();
	}
}
// }}}


// For a par proc: If closed, expand it and run discover on children to detect
// timers that appear directly after the PAR statement. Then return to select to
// choose one of those timers (if any, if no timers select will select the old
// timer, and serialization will eventually come back here. If the PAR is
// already expand, choose the next child to be serialized. 
// {{{ Proc* Scheduler::serializeParProc(...) 
Proc* Scheduler::serializeParProc(ParProc* par, SerializeResult* result) {
	if (par->getState() == CLOSED || par->getState() == EXPANDED) {
		doSerialize(par);
		discover(par, predictionTime);
		*result = RESELECT;
		return NULL;
	}
	else
		return par->getNextChild();
}
// }}}


// For a PAR end: If it is closed, but has got the number of parents as the
// opening PAR has children, then expand and return parEnd child as next proc
// to be run. If it is closed but has not enough parents, return original PAR.
// If it is not closed it has already been expanded, so just return child.
// {{{ Proc* Scheduler::serializeParEndProc(...) 
Proc* Scheduler::serializeParEndProc(ParEndProc* parEnd, SerializeResult* result) {
	if (parEnd->getState() == CLOSED) {
		if (parEnd->getNumParents() == parEnd->getPar()->countChildren()) {
			doSerialize(parEnd);
			return parEnd->getChild();
		}
		else
			return parEnd->getPar();
	}
	else
		return parEnd->getChild();
}
// }}}

// For a code proc: If the code is uncertain (does not know its execution
// time), set Scheduler::uncertain flag and return to select, which will then
// run everything up to this point as fast as possible. If there is not enough
// time in this horizon to add this proc, return to select. If all is well,
// add 
// {{{ Proc* Scheduler::serializeCodeProc(...) 
Proc* Scheduler::serializeCodeProc(CodeProc* code, int* horizon, SerializeResult* result) {
	if (code->isUncertain()) {
		*result = UNCERTAIN;
		return NULL;
	}

	int execTime = code->getExecTime();

	if (execTime > *horizon) {
		*result = COMPLETED;
		return NULL;	
	}

	if (code->getState() == CLOSED || code->getState() == EXPANDED) {
		code->setExecStart(predictionTime);
		doSerialize(code);
		predictionTime += execTime;
		*horizon -= execTime;
	}
	return code->getChild();
}
// }}}

// Serialize picks out the next timer to serialize, and calls serialize.
// Serialize may return for several reasons. One is RESELECT, that means new
// timers or choices appears. Then select picks a new timer and continues
// expanding with it. Another reason is UNCERTAIN, which means that some
// necessary proc cannot be expanded before all procs up to it are run. A
// third reason is COMPLETE, which means that expansion has completed up to the
// desired time horizon. 
// {{{ void Scheduler::select(...) 
void Scheduler::select(int horizon) {
	SerializeResult result = RESELECT;

	while (result == RESELECT) {
		// If there are no timers, return. The program has finished.
		if (timers.empty()) {
			printf("NO TIMERS\n");
			result = PROGRAMFINISHED;
			break;
		}

		print();

		// The timers are sorted on insertion.  If the timer with the shortest
		// deadline has an earliest start time in the future, do not process it
		// now. It must be processed as soon as it is ready, so this limits how
		// much time can be spent on other timers for now.
		if (timers.front()->getStartTime() > predictionTime) {
		// {{{
			printf("RESELECT   Timer in future\n");
			for (list<TimerProc*>::iterator it = timers.begin(); 
					it != timers.end(); ++it) {

				// First timer whose start time is not in the future. Call
				// serialize with horizon=time to ready time of earliest
				// deadline.
				if ((*it)->getStartTime() <= predictionTime) {
					serialize(*it, timers.front()->getStartTime() -
							predictionTime, &result);

					// if it returns COMPLETED, it has only completed the time
					// until the next timer is ready, and this means RESELECT.
					if (result == COMPLETED) {
						result = RESELECT;
					}
					break;
				}
			}

			// If there are no timers ready, insert sleep:
			predictionTime = timers.front()->getStartTime();
			result = RESELECT;
		}
		// }}}

		// If the timer with the shortest deadline is ready, then select it.
		else {
			serialize(timers.front(), horizon, &result);
		}

		printf("SELECT    STATUS: %s, PREDICTIONTIME:%d\n", 
				result == RESELECT ? "RESELECT" : 
				result == COMPLETED ? "COMPLETED" : 
				result == UNCERTAIN ? "UNCERTAIN" :
				result == OVERFLOW ? "OVERFLOW" : 
				result == PROGRAMFINISHED ? "PROGRAM FINISHED" : "",
				predictionTime);
	}

	assert(result != CONTINUE);
}
// }}}

// The discover() routing tries to look for timers that are not immediately
// available, but may be discovered by expanding control procs and other
// procs with no execution time associated with them. In other words, timers
// that starts in zero time after proc t. Discover is run recursively until it
// reaches non-control procs. Timers are added to the `timers` list, and their
// start-time is set to `time`. Do not worry if procs return NULL because they
// are uncertain, this will be picked up by select.
// {{{ void Scheduler::discover(...) 
void Scheduler::discover(Proc* t, int time) {
	// If NULL or not closed, the proc is already discovered
	if (t == NULL || t->getState() != CLOSED)
		return;

	TimerProc* timer;
	ControlProc* ctl;
	ParProc* par;
	SerializeResult dummy;

	// If the proc to be discovered is a timer, it is added to the timer list.
	// A timer is now found, but we need to dig deeper, in case there are nested
	// timers.
	if ((timer = dynamic_cast<TimerProc*>(t)) != NULL) {
		serializeTimerProc(timer, &dummy);
		discover(timer->getChild(), time);
	}

	// If the proc is a PAR proc, expand it and run discover on the
	// children.
	else if ((par = dynamic_cast<ParProc*>(t)) != NULL) {
		serializeParProc(par, &dummy);

		for (ProcIt it = par->getChildren().begin(); 
				it != par->getChildren().end(); ++it) {
			discover(*it, time);
		}
	}

	// If the proc is a control proc, expand it and run discover on the
	// child.
	else if ((ctl = dynamic_cast<ControlProc*>(t)) != NULL) {
		if (ctl->isDiscovery()) {
			serializeControlProc(ctl, &dummy);
			discover(ctl->getChild(), time);
		}
	}

}
// }}}

// {{{ void Scheduler::main(...) 
void Scheduler::main() {
	Proc* stop = add(new Stop());
	Proc* main = add(new Main(stop));

	// Run discover on main to discover startup timers.
	discover(main, now);
 
	printf("INITIAL DISCOVER COMPLETED\n");

	select(1000000);

	print();

	printf("\n\n");
	for (deque<Proc*>::iterator it = serialized.begin(); it != serialized.end(); ++it) {
		CodeProc* p = dynamic_cast<CodeProc*>(*it);
		if (p) {
			printf("T=%d\n", p->getExecStart());
			p->run();
		}
	}
}
// }}}

